/*-
 * %%----------------------------------------------------------------------------------------------
 * Archive ouverte UNIGE - AoU Resource Server Common - DownloadTokenService.java
 * SPDX-License-Identifier: GPL-2.0-or-later
 * %----------------------------------------------------------------------------------------------%
 * Copyright (C) 2022 - 2023 University of Geneva
 * %----------------------------------------------------------------------------------------------%
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as
 * published by the Free Software Foundation, either version 2 of the
 * License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public
 * License along with this program.  If not, see
 * <http://www.gnu.org/licenses/gpl-2.0.html>.
 * ----------------------------------------------------------------------------------------------%%
 */

package ch.unige.aou.service;

import java.security.SecureRandom;
import java.time.OffsetDateTime;
import java.time.format.DateTimeFormatter;
import java.util.Base64;
import java.util.concurrent.ThreadLocalRandom;

import org.springframework.http.HttpHeaders;
import org.springframework.scheduling.annotation.Scheduled;
import org.springframework.security.authentication.AnonymousAuthenticationToken;
import org.springframework.security.core.Authentication;
import org.springframework.security.core.context.SecurityContextHolder;
import org.springframework.stereotype.Service;
import org.springframework.web.bind.annotation.PathVariable;

import ch.unige.solidify.auth.model.UserInfo;
import ch.unige.solidify.service.HttpRequestInfoProvider;
import ch.unige.solidify.service.UserInfoService;
import ch.unige.solidify.util.HashTool;

import ch.unige.aou.config.AouProperties;
import ch.unige.aou.model.settings.DownloadToken;
import ch.unige.aou.repository.DownloadTokenRepository;

@Service
public class DownloadTokenService {

  private final UserInfoService userService;
  private final HttpRequestInfoProvider httpRequestInfoProvider;
  private final DownloadTokenRepository downloadTokenRepository;
  private final SecureRandom secureRandom;
  private final int downloadTokenLifeTimeMinutes;

  // Token length should be a multiple of three to avoid padding with '=' character that could be
  // problematic in cookie value
  private static final int DOWNLOAD_TOKEN_LENGTH = 201;
  private static final String DOWNLOAD_TOKEN_ANONYMOUS = "anonymous";

  private static final long CLEANING_INTERVAL_MILLISECONDS = 600000; // ten minutes

  public DownloadTokenService(UserInfoService userService, HttpRequestInfoProvider httpRequestInfoProvider,
          DownloadTokenRepository downloadTokenRepository, AouProperties aouProperties) {
    this.userService = userService;
    this.httpRequestInfoProvider = httpRequestInfoProvider;
    this.downloadTokenRepository = downloadTokenRepository;
    this.downloadTokenLifeTimeMinutes = aouProperties.getSecurity().getDownloadTokenLifeTimeMinutes();
    this.secureRandom = new SecureRandom();
  }

  @Scheduled(fixedDelay = CLEANING_INTERVAL_MILLISECONDS)
  public void removeExpiredDownloadToken() {
    final OffsetDateTime limitTime = OffsetDateTime.now().minusMinutes(this.downloadTokenLifeTimeMinutes);
    for (DownloadToken downloadToken : this.downloadTokenRepository.findExpiredDownloadTokens(limitTime)) {
      this.downloadTokenRepository.delete(downloadToken);
    }
  }

  public HttpHeaders getHeadersWithToken(@PathVariable String archiveId, String cookiePath) {
    String userId = DOWNLOAD_TOKEN_ANONYMOUS + ThreadLocalRandom.current().nextInt();
    final Authentication authentication = SecurityContextHolder.getContext().getAuthentication();
    if (authentication != null && !(authentication instanceof AnonymousAuthenticationToken)) {
      UserInfo userInfo = this.userService.findByExternalUid(authentication.getName());
      userId = userInfo.getResId();
    }
    // Create new download token
    final byte[] bytes = new byte[DOWNLOAD_TOKEN_LENGTH];
    this.secureRandom.nextBytes(bytes);
    final String token = Base64.getEncoder().encodeToString(bytes);
    final String tokenHash = HashTool.hash(token);
    DownloadToken downloadToken = new DownloadToken(userId, archiveId, tokenHash, token);

    synchronized (this) {
      // If there is already a download token for this resource and this user replace it
      DownloadToken existingToken = this.downloadTokenRepository.findByArchiveIdAndUserId(archiveId, userId);
      if (existingToken != null) {
        this.downloadTokenRepository.delete(existingToken);
      }

      this.downloadTokenRepository.save(downloadToken);
    }
    HttpHeaders headers = new HttpHeaders();
    OffsetDateTime expiresDate = OffsetDateTime.now().plusMinutes(this.downloadTokenLifeTimeMinutes);
    String expiresString = DateTimeFormatter.RFC_1123_DATE_TIME.format(expiresDate);
    headers.add(HttpHeaders.SET_COOKIE,
            "downloadToken-" + archiveId + "=" + token + "; Expires=" + expiresString + "; Path=" + cookiePath  + "; Secure; HttpOnly");
    return headers;
  }

}
